/********************************************************************
	created:	2007/05/04
	filename: 	Huffman.c
	file base:	Huffman
	file ext:	c
	author:		
	purpose:	إHuffman Code Table
*********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <memory.h>
#include "Huffman.h"

/**************************************************
//	Private
**************************************************/
typedef struct _TREENODE
{
	WORD index;
	DWORD freq;
	WORD level;
	struct _TREENODE *pParent, *pLeftChild, *pRightChild;
}TreeNode;

typedef struct _HUFFTREE
{
	WORD nCurTreeNodeNum;
	WORD nMaxTreeNodeNum;
	TreeNode **arrTreeNode;
	TreeNode *pRoot;
}HuffTree;

HuffTree* Build_HuffTree(HUFFMAN *pHuff);
bool Sort_Huff_TreeNode_By_Freq(HuffTree *pHuffTree, WORD Start, WORD End);
																//Direction=-1,+1k
bool Code_Shift(BYTE Code[MAX_CODE_LENGTH_BYTE], WORD length_bit, short Direction);
//////////////////////////////////////////////////////////////////////////

HUFFMAN* Create_Huffman(WORD nNodeNum)
{
	HUFFMAN *pHuff = NULL;
	int BufferSize = 0;
	if (nNodeNum <= 0) return NULL;

	BufferSize = 
		1 * sizeof(HUFFMAN)
		+ nNodeNum * sizeof(node)			//Huffman.Node_Table
		+ 1 * sizeof(HuffTree)				//pHuffTree
		+ 2 * nNodeNum * sizeof(TreeNode*)	//pHuffTree->arrTreeNode
		+ 2 * nNodeNum * sizeof(TreeNode);	//pHuffTree->arrTreeNode[]


	//tmbP@O餤
	//pHuff = (HUFFMAN*)malloc(BufferSize);
	pHuff = (HUFFMAN*)salloc(BufferSize);		//Barry 2008-06-27
	if (!pHuff) return NULL;
	memset(pHuff, 0, BufferSize);

	pHuff->NodeTable = (node*)(pHuff + 1);
	pHuff->_private = (void*)(pHuff->NodeTable + nNodeNum);
	pHuff->nNodeNum = nNodeNum;

	return pHuff;
}

bool Release_Huffman(HUFFMAN **ppHuff)
{
	if (!ppHuff || !*ppHuff) return false;
	//free(*ppHuff);	//Free@(Huffman*)
	sfree(*ppHuff);	//Free@(Huffman*)	//Barry 2008-06-02
	*ppHuff = NULL;
	return true;
}

bool Set_Huffman_Node(HUFFMAN *pHuff, WORD Index, DWORD Value, DWORD Frequency)
{
	if (!pHuff || !pHuff->NodeTable) return false;
	if (Index >= pHuff->nNodeNum) return false;

	memset(&(pHuff->NodeTable[Index]), 0, sizeof(node));
	pHuff->NodeTable[Index].index = Index;
	pHuff->NodeTable[Index].value = Value;
	pHuff->NodeTable[Index].freq = Frequency;

	return true;
}

bool Define_Huffman_Code(HUFFMAN *pHuff, WORD Index, WORD code_length_bit, BYTE Code[MAX_CODE_LENGTH_BYTE])
{
	if (!pHuff || !pHuff->NodeTable) return false;
	if (Index >= pHuff->nNodeNum || code_length_bit <= 0) return false;

	pHuff->NodeTable[Index].index = Index;
	pHuff->NodeTable[Index].code_length_bit = code_length_bit;
	memcpy(pHuff->NodeTable[Index].code, Code, MAX_CODE_LENGTH_BYTE);

	return true;
}

bool Read_Huffman_Table(HUFFMAN *pHuff)
{
	HuffTree *pHuffTree;	//VHuffman Tree
	TreeNode **TN;			//VTree Node Buffer
	TreeNode *pTempNode, *pNewNode;
	int Len;
	BYTE TempCode[MAX_CODE_LENGTH_BYTE];
	int i;
	if (!pHuff || !pHuff->NodeTable || !pHuff->_private) return false;

	//MHuffman TreeҦڪŶ
	memset(pHuff->_private, 0,
			+ 1 * sizeof(HuffTree)						//pHuffTree
			+ 2 * pHuff->nNodeNum * sizeof(TreeNode*)	//pHuffTree->arrTreeNode
			+ 2 * pHuff->nNodeNum * sizeof(TreeNode)	//pHuffTree->arrTreeNode[]
		  );

	//lTree Nodebuffer
	pHuffTree = (HuffTree*)pHuff->_private;
	pHuffTree->nMaxTreeNodeNum = pHuff->nNodeNum * 2;
	pHuffTree->arrTreeNode = (TreeNode**)(pHuffTree + 1);
	pTempNode = (TreeNode*)(pHuffTree->arrTreeNode + pHuffTree->nMaxTreeNodeNum);
	TN = pHuffTree->arrTreeNode;	//TNNTree Node Buffer Array
	for (i=0; i<pHuffTree->nMaxTreeNodeNum; i++)
	{
		TN[i] = pTempNode++;
	}

	//إRoot Node
	pHuffTree->pRoot = TN[0];
	pHuffTree->nCurTreeNodeNum++;

	//QNodeTableCode WordإHuffman Tree
	for (i=0; i<pHuff->nNodeNum; i++)
	{
		Len = pHuff->NodeTable[i].code_length_bit;	//Code
		if (Len <= 0) continue;						//ˬdCode

		memcpy(TempCode, pHuff->NodeTable[i].code, MAX_CODE_LENGTH_BYTE);	//Code Word

		pTempNode = pHuffTree->pRoot;
		while (Len >= 0)		//Top-DownإHuffman Tree
		{
			if (Len > pTempNode->level)		//slevel
			{
				pTempNode->level = Len;
			}
			if (Len > 0)					//Internal Nodeindex
			{
				pTempNode->index = -1;
			}
			else	//Len = 0, wLeaf Node, X
			{
				pTempNode->index = pHuff->NodeTable[i].index;
				break;
			}

			if (TempCode[0] & 0x80)	//kChild
			{
				if (!pTempNode->pRightChild)	//sbNsW@Node
				{
					pTempNode->pRightChild = TN[pHuffTree->nCurTreeNodeNum];
					pHuffTree->nCurTreeNodeNum++;
				}
				pNewNode = pTempNode->pRightChild;
			}
			else	//Child
			{
				if (!pTempNode->pLeftChild)	//sbNsW@Node
				{
					pTempNode->pLeftChild = TN[pHuffTree->nCurTreeNodeNum];
					pHuffTree->nCurTreeNodeNum++;
				}
				pNewNode = pTempNode->pLeftChild;
			}
			pNewNode->pParent = pTempNode;

			Code_Shift(TempCode, (WORD)Len, -1);	//Code WordVShift@
			Len--;							//GCodeפ]1

			pTempNode = pNewNode;
		}
	}

	return true;
}


BYTE bit_mask[8] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
bool Huffman_Decode_Once(HUFFMAN *pHuff, BYTE *Bitstring, WORD StartBit, WORD* pRetIndex, WORD* pRetLenBit)
{
	BYTE *pCurByte;
	int CurBitPos;
	int Len, Index;
	HuffTree *pHuffTree;	//VHuffman Tree
	TreeNode *pTempNode;

	if (!pHuff || !pHuff->NodeTable || !pHuff->_private) 
	{
		return false;
	}

	if (!Bitstring || !pRetIndex || !pRetLenBit)
	{
		return false;
	}

	pHuffTree = (HuffTree*)pHuff->_private;	//Huffman Tree
	if (!pHuffTree->arrTreeNode || !pHuffTree->pRoot)
	{
		return false;
	}

	pCurByte = Bitstring + (StartBit  >> 3);		//Bitstring + StartBit /8
	CurBitPos = (StartBit & 0x07);				//StartBit % 8

	Index = -1;
	pTempNode = pHuffTree->pRoot;
	Len = pHuffTree->pRoot->level;		//Code Word̪uRootLevel
	while (Len > 0)
	{
		if (*pCurByte & bit_mask[CurBitPos])	//kChild:1
		{
			pTempNode = pTempNode->pRightChild;
		}
		else					//Child:0
		{
			pTempNode = pTempNode->pLeftChild;
		}

		if (!pTempNode) break;
		
		if (pTempNode->level == 0)
		{
			Index = pTempNode->index;
			break;
		}

		if ((++CurBitPos) == 8)
		{
			CurBitPos = 0;
			pCurByte++;
		}
		Len--;
	}	

	if (Index == -1)
	{
		return false;
	}
	else
	{
		*pRetIndex = pHuff->NodeTable[Index].index;
		*pRetLenBit = pHuff->NodeTable[Index].code_length_bit;
	}

	return true;
}

bool Huffman_Decode_Multiple_Bytes(HUFFMAN *pHuff, BYTE *Bitstring, WORD StartBit, WORD nDecodeNum, BYTE *OutBuf)
{
	BYTE *pCurByte;
	int CurBitPos;
	int Len, Index;
	HuffTree *pHuffTree;	//VHuffman Tree
	TreeNode *pTempNode;
	int i, k;
	BYTE bitmask, bytevalue;
	
	if (!pHuff || !pHuff->NodeTable || !pHuff->_private) 
	{
		return false;
	}

	if (!Bitstring)
	{
		return false;
	}

	pHuffTree = (HuffTree*)pHuff->_private;	//Huffman Tree
	if (!pHuffTree->arrTreeNode || !pHuffTree->pRoot)
	{
		return false;
	}

	pCurByte = Bitstring + (StartBit>>3);	//Bitstring + StartBit /8
	CurBitPos = (StartBit & 0x0007);		//StartBit % 8
	
	for (i=0; i<nDecodeNum; i++)
	{
		Index = -1;
		pTempNode = pHuffTree->pRoot;
		Len = pHuffTree->pRoot->level;		//Code Word̪uRootLevel
		while (Len > 0)
		{
			if (*pCurByte & bit_mask[CurBitPos])	//kChild:1
			{
				pTempNode = pTempNode->pRightChild;
			}
			else					//Child:0
			{
				pTempNode = pTempNode->pLeftChild;
			}

			if (!pTempNode) break;
			
			if ((++CurBitPos) == 8)
			{
				CurBitPos = 0;
				pCurByte++;
			}

			if (pTempNode->level == 0)
			{
				Index = pTempNode->index;
				break;
			}
			Len--;
		}	

		if (Index == -1)
		{
			return false;
		}
		else
		{
			bytevalue = (BYTE)(pHuff->NodeTable[Index].value);
			//JImageBuffer
			bitmask = 0x80;
			for (k=0; k<8; k++)
			{
				if (bytevalue & bitmask)
				{
					*OutBuf++ = 1;
				}
				else
				{
					*OutBuf++ = 0;
				}
				bitmask >>= 1;
			}
		}
	}

	return true;
}

															//-1:Left, 1:Right
bool Code_Shift(BYTE Code[MAX_CODE_LENGTH_BYTE], WORD length_bit, short Direction)
{
	int length_byte;			//ݭnBYTE
	BYTE res_left, res_right;	//Shift|vTU@BYTE
	int i;
	
	if (Direction != -1 && Direction != 1)
	{
		return false;
	}
	if (length_bit > MAX_CODE_LENGTH_BYTE * 8)
	{
		return false;
	}

	length_byte = (length_bit+7) / 8;
	res_left = 0;
	res_right = 0;
	if (Direction == -1)	//VShift
	{
		for (i=length_byte-1; i>=0; i--)
		{
			res_right = res_left;					//ŪJkiLӪƦr
			res_left = ((Code[i]&0x80) ? 1 : 0);	//ݬO_ݶi1쥪䪺BYTE
			Code[i] <<= 1;
			if (res_right)
			{
				Code[i] |= 0x01;					//[WqkiLӪ
			}
			
		}
	}
	else	//Direction = 1 VkShift
	{
		for (i=0; i<length_byte; i++)
		{
			res_left = res_right;					//ŪJhLӪƦr
			res_right = ((Code[i]&0x01) ? 1 : 0);	//ݬO_ݰh1k䪺BYTE
			Code[i] >>= 1;
			if (res_left)
			{
				Code[i] |= 0x80;					//[WqhLӪ
			}
		}
	}

	return true;
}																


