
#include "config.h"

#include <stdio.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <stdlib.h>
#include <unistd.h>

#include "mp_msg.h"
#include "help_mp.h"

#include "osdep/timer.h"
#include "osdep/shmem.h"

#include "stream.h"
#include "demuxer.h"
#include "parse_es.h"

#include "codec-cfg.h"

#include "libvo/video_out.h"

#include "stheader.h"
#include "vd.h"
#include "vf.h"

#include "dec_video.h"

#ifdef DYNAMIC_PLUGINS
#include <dlfcn.h>
#endif


// ===================================================================
#if defined __ARC_MPLAYER__
static int disable_timer_update = 1;
#endif

extern unsigned int video_time_usage;
extern unsigned int vout_time_usage;

#include "cpudetect.h"

static int framesProduced = 0;
int divx_quality=0;

vd_functions_t* mpvdec=NULL;

//for scalar Barry 2007-05-11
int display_frame = 0;

int get_video_quality_max(sh_video_t *sh_video){
  vf_instance_t* vf=sh_video->vfilter;
  if(vf){
    int ret=vf->control(vf,VFCTRL_QUERY_MAX_PP_LEVEL,NULL);
    if(ret>0){
      mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_UsingExternalPP,ret);
      return ret;
    }
  }
  if(mpvdec){
    int ret=mpvdec->control(sh_video,VDCTRL_QUERY_MAX_PP_LEVEL,NULL);
    if(ret>0){
      mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_UsingCodecPP,ret);
      return ret;
    }
  }
//  mp_msg(MSGT_DECVIDEO,MSGL_INFO,"[PP] Sorry, postprocessing is not available\n");
  return 0;
}

void set_video_quality(sh_video_t *sh_video,int quality){
  vf_instance_t* vf=sh_video->vfilter;
  if(vf){
    int ret=vf->control(vf,VFCTRL_SET_PP_LEVEL, (void*)(&quality));
    if(ret==CONTROL_TRUE) return; // success
  }
  if(mpvdec)
    mpvdec->control(sh_video,VDCTRL_SET_PP_LEVEL, (void*)(&quality));
}

int set_video_colors(sh_video_t *sh_video,char *item,int value)
{
    vf_instance_t* vf=sh_video->vfilter;
    vf_equalizer_t data;

    data.item = item;
    data.value = value;

    mp_dbg(MSGT_DECVIDEO,MSGL_V,"set video colors %s=%d \n", item, value);
    if (vf)
    {
	int ret = vf->control(vf, VFCTRL_SET_EQUALIZER, &data);
	if (ret == CONTROL_TRUE)
	    return(1);
    }
    /* try software control */
    if(mpvdec)
	if( mpvdec->control(sh_video,VDCTRL_SET_EQUALIZER, item, (int *)value)
	    == CONTROL_OK) return 1;
    mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_VideoAttributeNotSupportedByVO_VD,item);
    return 0;
}

int get_video_colors(sh_video_t *sh_video,char *item,int *value)
{
    vf_instance_t* vf=sh_video->vfilter;
    vf_equalizer_t data;

    data.item = item;

    mp_dbg(MSGT_DECVIDEO,MSGL_V,"get video colors %s \n", item);
    if (vf)
    {
        int ret = vf->control(vf, VFCTRL_GET_EQUALIZER, &data);
	if (ret == CONTROL_TRUE){
	    *value = data.value;
	    return(1);
	}
    }
    /* try software control */
    if(mpvdec) return mpvdec->control(sh_video,VDCTRL_GET_EQUALIZER, item, value);
    return 0;
}

int set_rectangle(sh_video_t *sh_video,int param,int value)
{
    vf_instance_t* vf=sh_video->vfilter;
    int data[] = {param, value};

    mp_dbg(MSGT_DECVIDEO,MSGL_V,"set rectangle \n");
    if (vf)
    {
        int ret = vf->control(vf, VFCTRL_CHANGE_RECTANGLE, data);
	if (ret)
	    return(1);
    }
    return 0;
}

void resync_video_stream(sh_video_t *sh_video)
{
#if defined __ARC_MPLAYER__
  disable_timer_update = 1;
#endif
    if(mpvdec) mpvdec->control(sh_video, VDCTRL_RESYNC_STREAM, NULL);
}

void uninit_video(sh_video_t *sh_video){
    if(!sh_video->inited) return;
    mp_msg(MSGT_DECVIDEO,MSGL_V,MSGTR_UninitVideoStr,sh_video->codec->drv);
    mpvdec->uninit(sh_video);
#ifdef DYNAMIC_PLUGINS
    if (sh_video->dec_handle)
	dlclose(sh_video->dec_handle);
#endif
    vf_uninit_filter_chain(sh_video->vfilter);
    sh_video->inited=0;
}

void vfm_help(){
    int i;
    mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_AvailableVideoFm);
    mp_msg(MSGT_DECVIDEO,MSGL_INFO,"   vfm:    info:  (comment)\n");
    for (i=0; mpcodecs_vd_drivers[i] != NULL; i++)
	mp_msg(MSGT_DECVIDEO,MSGL_INFO,"%8s  %s (%s)\n",
	    mpcodecs_vd_drivers[i]->info->short_name,
	    mpcodecs_vd_drivers[i]->info->name,
	    mpcodecs_vd_drivers[i]->info->comment);
}

int init_video(sh_video_t *sh_video,char* codecname,char* vfm,int status){
    unsigned int orig_fourcc=sh_video->bih?sh_video->bih->biCompression:0;
    sh_video->codec=NULL;
    sh_video->vf_inited=0;

    while(1){
	int i;
	// restore original fourcc:
	if(sh_video->bih) sh_video->bih->biCompression=orig_fourcc;
	if(!(sh_video->codec=find_codec(sh_video->format,
          sh_video->bih?((unsigned int*) &sh_video->bih->biCompression):NULL,
          sh_video->codec,0) )) break;
	// ok we found one codec
	if(sh_video->codec->flags&CODECS_FLAG_SELECTED) continue; // already tried & failed
	if(codecname && strcmp(sh_video->codec->name,codecname)) continue; // -vc
	if(vfm && strcmp(sh_video->codec->drv,vfm)) continue; // vfm doesn't match
	if(sh_video->codec->status<status) continue; // too unstable
	sh_video->codec->flags|=CODECS_FLAG_SELECTED; // tagging it
	// ok, it matches all rules, let's find the driver!
	for (i=0; mpcodecs_vd_drivers[i] != NULL; i++)
//	    if(mpcodecs_vd_drivers[i]->info->id==sh_video->codec->driver) break;
	    if(!strcmp(mpcodecs_vd_drivers[i]->info->short_name,sh_video->codec->drv)) break;
	mpvdec=mpcodecs_vd_drivers[i];
#ifdef DYNAMIC_PLUGINS
	if (!mpvdec)
	{
	    /* try to open shared decoder plugin */
	    int buf_len;
	    char *buf;
	    vd_functions_t *funcs_sym;
	    vd_info_t *info_sym;

	    buf_len = strlen(MPLAYER_LIBDIR)+strlen(sh_video->codec->drv)+16;
	    buf = malloc(buf_len);
	    if (!buf)
		break;
	    snprintf(buf, buf_len, "%s/mplayer/vd_%s.so", MPLAYER_LIBDIR, sh_video->codec->drv);
	    mp_msg(MSGT_DECVIDEO, MSGL_DBG2, "Trying to open external plugin: %s\n", buf);
	    sh_video->dec_handle = dlopen(buf, RTLD_LAZY);
	    if (!sh_video->dec_handle)
		break;
	    snprintf(buf, buf_len, "mpcodecs_vd_%s", sh_video->codec->drv);
	    funcs_sym = dlsym(sh_video->dec_handle, buf);
	    if (!funcs_sym || !funcs_sym->info || !funcs_sym->init ||
		!funcs_sym->uninit || !funcs_sym->control || !funcs_sym->decode)
		break;
	    info_sym = funcs_sym->info;
	    if (strcmp(info_sym->short_name, sh_video->codec->drv))
		break;
	    free(buf);
	    mpvdec = funcs_sym;
	    mp_msg(MSGT_DECVIDEO, MSGL_V, "Using external decoder plugin (%s/mplayer/vd_%s.so)!\n",
		MPLAYER_LIBDIR, sh_video->codec->drv);
	}
#endif
	if(!mpvdec){ // driver not available (==compiled in)
	    mp_msg(MSGT_DECVIDEO,MSGL_WARN,MSGTR_VideoCodecFamilyNotAvailableStr,
		sh_video->codec->name, sh_video->codec->drv);
	    continue;
	}
	// it's available, let's try to init!
	if(sh_video->codec->flags & CODECS_FLAG_ALIGN16){
	    // align width/height to n*16
	    // FIXME: save orig w/h, and restore if codec init failed!
	    if(sh_video->bih){
		sh_video->disp_w=sh_video->bih->biWidth=(sh_video->bih->biWidth+15)&(~15);
		sh_video->disp_h=sh_video->bih->biHeight=(sh_video->bih->biHeight+15)&(~15);
	    } else {
		sh_video->disp_w=(sh_video->disp_w+15)&(~15);
		sh_video->disp_h=(sh_video->disp_h+15)&(~15);
	    }
	}
	// init()
	mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_OpeningVideoDecoder,mpvdec->info->short_name,mpvdec->info->name);
	if(!mpvdec->init(sh_video)){
	    mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_VDecoderInitFailed);
	    continue; // try next...
	}
	// Yeah! We got it!
	sh_video->inited=1;
	return 1;
    }
    return 0;
}

extern char *get_path(char *filename);

int init_best_video_codec(sh_video_t *sh_video,char** video_codec_list,char** video_fm_list){
char* vc_l_default[2]={"",(char*)NULL};
// hack:
if(!video_codec_list) video_codec_list=vc_l_default;
// Go through the codec.conf and find the best codec...
sh_video->inited=0;
codecs_reset_selection(0);
while(!sh_video->inited && *video_codec_list){
  char* video_codec=*(video_codec_list++);
  if(video_codec[0]){
    if(video_codec[0]=='-'){
      // disable this codec:
      select_codec(video_codec+1,0);
    } else {
      // forced codec by name:
      mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_ForcedVideoCodec,video_codec);
      init_video(sh_video,video_codec,NULL,-1);
    }
  } else {
    int status;
    // try in stability order: UNTESTED, WORKING, BUGGY. never try CRASHING.
    if(video_fm_list){
      char** fmlist=video_fm_list;
      // try first the preferred codec families:
      while(!sh_video->inited && *fmlist){
        char* video_fm=*(fmlist++);
	mp_msg(MSGT_DECVIDEO,MSGL_INFO,MSGTR_TryForceVideoFmtStr,video_fm);
	for(status=CODECS_STATUS__MAX;status>=CODECS_STATUS__MIN;--status)
	    if(init_video(sh_video,NULL,video_fm,status)) break;
      }
    }
    if(!sh_video->inited)
	for(status=CODECS_STATUS__MAX;status>=CODECS_STATUS__MIN;--status)
	    if(init_video(sh_video,NULL,NULL,status)) break;
  }
}

if(!sh_video->inited){
    mp_msg(MSGT_DECVIDEO,MSGL_ERR,MSGTR_CantFindVideoCodec,sh_video->format);
    mp_msg(MSGT_DECAUDIO,MSGL_HINT, MSGTR_RTFMCodecs);
    return 0; // failed
}

mp_msg(MSGT_DECVIDEO,MSGL_INFO,"Selected video codec: [%s] vfm:%s (%s)\n",
    sh_video->codec->name,sh_video->codec->drv,sh_video->codec->info);
return 1; // success
}

extern int vo_directrendering;

extern int preview_frame;	// Raymond 2007/05/15

extern int is_vc1;		// Raymond 2007/08/15

extern unsigned char *preview_buf;		// Raymond 2008/06/23

// Raymond
int scale_w;
int scale_h;

#ifdef BUILD_SK2851	// Raymond 2008/08/20
extern int set_real_display;
extern int disp_x; 
extern int disp_y;
#endif

typedef struct 
{
	unsigned char	*y_buf;
	unsigned char	*u_buf;
	unsigned char	*v_buf;
	int width;
	int height;
} YUV_frame;

#define OUTPUT_XRES		1920
#define SCALE_SHIFT		16

static unsigned char ScaleBufY[OUTPUT_XRES];
static unsigned char ScaleBufU[OUTPUT_XRES];
static unsigned char ScaleBufV[OUTPUT_XRES];

//#define ScaleColor(col0, col1, part,  all, div)	( (((col1*part) + (col0*(all-part)))*div) >> SCALE_SHIFT)
#define ScaleColor(col0, col1, part,  all, div)	col0		//Barry 2007-12-07

// Height can not extend !!
void Scale(YUV_frame *input, YUV_frame *scale)
{
	int w = input->width;
	int h = input->height;
	int sw = scale->width;
	int sh = scale->height;
	int xdiv = ( 1 << SCALE_SHIFT ) / sw;
	int ydiv = ( 1 << SCALE_SHIFT ) / sh;
	int i, j, k, m = 0;

	int ybase, ubase, vbase;
	int partx = 0, sidx, scaley;
	int party = 0, yline = 0;
	int sw2 = sw >> 1;
	int sh2 = sh >> 1;

	unsigned char *pY = input->y_buf;
	unsigned char *pU = input->u_buf;
	unsigned char *pV = input->v_buf;

	unsigned char *sY = scale->y_buf;
	unsigned char *sU = scale->u_buf;
	unsigned char *sV = scale->v_buf;
	
	// Raymond	
	if( w == sw && h == sh )
	{	
		for( i = 0 ; i < sh ; i++ )
			memcpy(sY + i * sw, pY + i * w, sw);
		
		w >>= 1;
		h >>= 1;
		sw >>= 1;
		sh >>= 1;
		
		for( i = 0 ; i < sh ; i++ )
			memcpy(sU + i * sw, pU + i * w, sw);
		for( i = 0 ; i < sh ; i++ )
			memcpy(sV + i * sw, pV + i * w, sw);
		
		return;
	}	

//
if(sh > h)
	for( j = 0 ; j < sh ; j++)
	{
		ybase = sw * j;
		ubase = ybase >> 1;
		vbase = ubase;
		
		partx = sidx = scaley = 0;

		party += h;
		while ( party >= sh )
		{
//			scaley = party+1;
			party -= sh;

			pY += w;
			pU += w >> 1;
			pV += w >> 1;
		}

		for ( i = k = 0 ; i < sw ; i++, k++ )
		{
			unsigned char y, u, v;
			partx += w;
			while ( partx >= sw )
			{
				sidx ++;
				partx -= sw;
			}
			
			y = ScaleColor(pY[sidx-1], pY[sidx], partx, sw, xdiv);	
			if( j == 0 )
				ScaleBufY[k] = y;
			
			sY[ybase] = ScaleColor(ScaleBufY[k], y, party, sh, ydiv);
			ybase++;
			ScaleBufY[k] = y;
			
			if( j < sh2 && i < sw2 ) // UV
			{
				u = ScaleColor(pU[sidx-1], pU[sidx], partx, sw, xdiv);
				v = ScaleColor(pV[sidx-1], pV[sidx], partx, sw, xdiv);	
				if( j == 0 )
				{
					ScaleBufU[k] = u;
					ScaleBufV[k] = v;
				}
				sU[ubase] = ScaleColor(ScaleBufU[k], u, party, sh, ydiv);
				sV[vbase] = ScaleColor(ScaleBufV[k], v, party, sh, ydiv);
				ubase++;
				vbase++;
				
				ScaleBufU[k] = u;
				ScaleBufV[k] = v;				
			}
		}
	}
else
	for( j = 0 ; j < h ; j++)
	{
		ybase = sw * yline;
		ubase = ybase >> 1;
		vbase = ubase;
		
		partx = sidx = scaley = 0;
		if (party >= sh) 
		{
			party -= sh; 
			if (party < sh)
			{
				scaley = party+1;
				yline++;
				party += h;
			}
		}
		else
			party += h;
		
		for ( i = k = 0 ; i < sw ; i++, k++ )
		{
			unsigned char y, u, v;
			partx += w;
			while ( partx >= sw )
			{
				sidx ++;
				partx -= sw;
			}
			
			y = ScaleColor(pY[sidx-1], pY[sidx], partx, sw, xdiv);
			if (scaley) 
			{
				sY[ybase++] = ScaleColor(ScaleBufY[k], y, scaley, sh, ydiv);
			}
			
			ScaleBufY[k] = y;
						
			if( m < sh2 && i < sw2 ) // UV	
			{
				u = ScaleColor(pU[sidx-1], pU[sidx], partx, sw, xdiv);
				v = ScaleColor(pV[sidx-1], pV[sidx], partx, sw, xdiv);
				
				if (scaley) 
				{
					sU[ubase++] = ScaleColor(ScaleBufU[k], u, scaley, sh, ydiv);
					sV[vbase++] = ScaleColor(ScaleBufV[k], v, scaley, sh, ydiv);
				}
				
				ScaleBufU[k] = u;
				ScaleBufV[k] = v;
			}
		}

		if(scaley)
			m++;

		pY += w;
		pU += w >> 1;
		pV += w >> 1;
	}

}

#include "pgu_lib.h"

#define range_limit(x) ( (x > 255) ? 255 : ( ( x < 0 ) ? 0 : x ) )

// faster YUV420_to_RGB555

#define MUL_1402		1436
#define MUL_034414		-352
#define MUL_071414		-731
#define MUL_1772		1815
#define SCALE_FACTOR	10
/*
r = y + (int)(cr * 1.402);
g = y + (int)( (cb * -0.34414) + (cr * -0.71414) );
b = y + (int)(cb * 1.772);
*/

void YUV420_to_RGB555(YUV_frame *scale, unsigned short *rgb_buffer)
{
	unsigned char *sY = scale->y_buf;
	unsigned char *sU = scale->u_buf;
	unsigned char *sV = scale->v_buf;
	unsigned short *outptr = rgb_buffer;

	int sw = scale->width;
	int sh = scale->height;
	int sw2 = sw >> 1;
	int sh2 = sh >> 1;
	
	int y00, y01, y10, y11;
	int cb, cr;
	int cr1, cb1, cbcr;
	int r, g, b;
	int i, j, i2;

	for( j = 0 ; j < sh2 ; j++ )
	{
		for( i = 0 ; i < sw2 ; i++ )
		{
			i2 = i << 1;
			y00  = sY[i2];
			y01  = sY[i2+1];
			y10  = sY[i2+sw];
			y11  = sY[i2+sw+1];

			cb = sU[i] - 128;
			cr = sV[i] - 128;

			cr1  = (cr * MUL_1402) >> SCALE_FACTOR;
			cb1  = (cb * MUL_1772) >> SCALE_FACTOR;
			cbcr = ( (cb * MUL_034414) + (cr * MUL_071414) ) >> SCALE_FACTOR;

			/* Range-limiting is essential due to noise introduced by DCT losses. */

			r = range_limit(y00 + cr1);
			g = range_limit(y00 + cbcr);
			b = range_limit(y00 + cb1);
			outptr[0] = ENCODE_RGB555(r, g, b);

			r = range_limit(y01 + cr1);
			g = range_limit(y01 + cbcr);
			b = range_limit(y01 + cb1);
			outptr[1] = ENCODE_RGB555(r, g, b);
				
			r = range_limit(y10 + cr1);
			g = range_limit(y10 + cbcr);
			b = range_limit(y10 + cb1);
			outptr[sw] = ENCODE_RGB555(r, g, b);

			r = range_limit(y11 + cr1);
			g = range_limit(y11 + cbcr);
			b = range_limit(y11 + cb1);
			outptr[sw+1] = ENCODE_RGB555(r, g, b);

			outptr += 2 ;
		}

		outptr += sw ;
		sY += sw << 1;
		sU += sw2;
		sV += sw2;
	}
}

int preview_video(sh_video_t *sh_video,unsigned char *start,int in_size,int drop_frame)
{
	mp_image_t *mpi=NULL;
	int ret = 0;

	mpi = mpvdec->decode(sh_video, start, in_size, drop_frame);
			  
 	if( mpi != NULL )	  
	{	
		static int cnt = 0;
		int i = 0;
		int sum = 0;
		unsigned char	*scale_buffer = NULL;
		unsigned short* framebuf;
		int scale_size = scale_w * scale_h;

		//Barry 2007-05-15
		if( strcmp(mpvdec->info->short_name,"msvc1") )
              	if (mpi->w & 0xf)
                     	mpi->w = ((mpi->w + 0xf) & ~0xf);

		int image_size = mpi->w * mpi->h;
		
		YUV_frame input;
		YUV_frame scale;

		input.y_buf = (unsigned char *)mpi->planes[0];		
#if 1	
		cnt++;

		if( cnt < preview_frame )	// Raymond 2007/05/15
		{
			// Raymond 2007/05/08
			for( i = 0 ; i < image_size ; i++)
				sum += input.y_buf[i]; 

			sum = sum / image_size; 
			
			printf("Image[%d] Y = %d\n", cnt, sum);

			if( sum < 32 )
				return 0;
		}
/*		
		// for debug
		for( i = 0 ; i < image_size ; i++)
			sum += input.y_buf[i]; 

		sum = sum / image_size; 
*/		
#endif

//		printf("Input WxH = %3d x %3d => Scale WxH = %3d x %3d\n", mpi->w, mpi->h, scale_w, scale_h);		
		
#if 1	// Raymond 2007/03/12
//		if( !strcmp(mpvdec->info->short_name,"msvc1") )	
		if( is_vc1 == 1 && ( (mpi->w & 0xf) || (mpi->h & 0xf) ) )	// Raymond 2008/07/15
		{
			// VC-1 : swap UV 
			input.u_buf = (unsigned char *)mpi->planes[2];
			input.v_buf = (unsigned char *)mpi->planes[1];
		}
		else
		{
			input.u_buf = (unsigned char *)mpi->planes[1];
			input.v_buf = (unsigned char *)mpi->planes[2];
		}
#else
		input.u_buf = (unsigned char *)mpi->planes[1];
		input.v_buf = (unsigned char *)mpi->planes[2];
#endif		
		input.width = mpi->w;
		input.height = mpi->h;

#ifndef BUILD_SK2851		// Raymond 2008/06/23
		// open 
		pgu_open_fb();  
		framebuf = (unsigned short*)pgu_map_buffer(scale_size * 4, 0);
#else
		framebuf = preview_buf;
//		printf("framebuf = %X\n", framebuf);

		// Raymond 2008/08/20
		if( set_real_display == 1 )
		{
			int i = 0;
			int iw = input.width;
			int iw2 = input.width >> 1;
			int rh2  = disp_y >> 1;
			int dw2 = disp_x >> 1;			

			// Y		
			for( i = 0 ; i < disp_y ; i++ )
				memcpy(input.y_buf+i*disp_x, input.y_buf+i*iw, disp_x);
			
			// U		
			for( i = 0 ; i < rh2; i++ )
				memcpy(input.u_buf+i*dw2, input.u_buf+i*iw2, dw2);

			// V		
			for( i = 0 ; i < rh2; i++ )
				memcpy(input.v_buf+i*dw2, input.v_buf+i*iw2, dw2);

			input.width = disp_x;
			input.height = disp_y;
			
		}

#endif		

		scale_buffer = (unsigned char *)framebuf + scale_size * 2;	
		scale.y_buf = scale_buffer;
		scale.u_buf = scale.y_buf + scale_size;
		scale.v_buf = scale.u_buf + scale_size / 4 ;
		scale.width = scale_w;
		scale.height = scale_h;
			
		// scale mpi->w x mpi->h to scale_w x scale_h
		Scale(&input, &scale);
#if 0			
		{
			FILE *yuv_out = fopen("out.yuv","wb");
			if( yuv_out == NULL )
				return 0;
			fwrite( (unsigned char *)scale.y_buf, 1, scale_size, yuv_out);
			fwrite( (unsigned char *)scale.u_buf, 1, scale_size/4, yuv_out);
			fwrite( (unsigned char *)scale.v_buf, 1, scale_size/4, yuv_out);
			fclose(yuv_out);
		}
#endif		
		// convert to RGB555
		YUV420_to_RGB555(&scale, framebuf);

#ifndef BUILD_SK2851		// Raymond 2008/06/23
		// close
		pgu_close_fb();
#endif

//		printf("Preview WxH = %3d x %3d => %3d x %3d", mpi->w, mpi->h, scale_w, scale_h);

		printf("Preview %3dx%3d - Frame %d (Y = %d)", mpi->w, mpi->h, cnt, sum);
//		printf("Preview %3dx%3d - #%d", mpi->w, mpi->h, cnt);
				
		return 1;
	} 
	return ret;
}

int getyuv = 0;	// Raymond

int decode_video(sh_video_t *sh_video,unsigned char *start,int in_size,int drop_frame){
vf_instance_t* vf;
mp_image_t *mpi=NULL;
unsigned int t=GetTimer();
unsigned int t2;
//double tt;
int ret;

//if(!(sh_video->ds->flags&1) || sh_video->ds->pack_no<5)
  mpi=mpvdec->decode(sh_video, start, in_size, drop_frame);
//------------------------ frame decoded. --------------------
	
	// Raymond
	if(getyuv == 1)
	{
		if( mpi != NULL )	  
		{
			int image_size = mpi->w * mpi->h;
		
			FILE *yuv_out = fopen("out.yuv","ab");
			if( yuv_out == NULL )
				return 0;
			fwrite( (unsigned char *)mpi->planes[0], 1, image_size, yuv_out);
			fwrite( (unsigned char *)mpi->planes[1], 1, image_size/4, yuv_out);
			fwrite( (unsigned char *)mpi->planes[2], 1, image_size/4, yuv_out);
			fclose(yuv_out);

			printf("getyuv = %d : %3dx%3d\n", getyuv, mpi->w, mpi->h);
		}	
	}

#ifdef ARCH_X86
	// some codecs are broken, and doesn't restore MMX state :(
	// it happens usually with broken/damaged files.
if(gCpuCaps.has3DNow){
	__asm __volatile ("femms\n\t":::"memory");
}
else if(gCpuCaps.hasMMX){
	__asm __volatile ("emms\n\t":::"memory");
}
#endif

t2=GetTimer();
t=t2-t;
//tt = t*0.000001f;

#ifdef __ARC_MPLAYER__
//keep track of whether any frames were produced by this call to decode
if (mpi != NULL)
{
  framesProduced = 1;
  //fprintf(stderr," framesProduced");
  sh_video->num_frames_decoded++;
  sh_video->num_frames++;
  display_frame = sh_video->num_frames;		//for scalar Barry 2007-05-11
  disable_timer_update = 0;  
}
else
  framesProduced = 0;
if (!disable_timer_update)
//  video_time_usage += tt;
  video_time_usage += t;
#else
//video_time_usage += tt;
video_time_usage += t;
#endif

#if 0
if(!mpi || drop_frame) return 0; // error / skipped frame
#else	// Raymond 2007/02/15 - only drop frame for libarc

if(!mpi) 
	return 0; // error 

// Raymond 2007/08/23 - disable
//if( !strcmp(mpvdec->info->short_name,"libarc") && drop_frame) 
//	return 0; // dropped frame

#endif

#if 1	// Raymond 2008/01/03
	if( is_vc1 == 1 )
	{
		vf=sh_video->vfilter;
		ret = vf->put_image(vf,mpi); // apply video filters and call the leaf vo/ve
		if(ret>0) vf->control(vf,VFCTRL_DRAW_OSD,NULL);
	    	t2=GetTimer()-t2;
//	    	tt=t2*0.000001f;
//	    	vout_time_usage+=tt;
		vout_time_usage+=t2;
	}
	else
	{
		free(mpi);
		ret = 1;
	}
#else
//vo_draw_image(video_out,mpi);
vf=sh_video->vfilter;
ret = vf->put_image(vf,mpi); // apply video filters and call the leaf vo/ve
if(ret>0) vf->control(vf,VFCTRL_DRAW_OSD,NULL);
    t2=GetTimer()-t2;
    tt=t2*0.000001f;
    vout_time_usage+=tt;

// Raymond 2007/06/28 - fix memory leakage bug
//if(mpi) 
if( mpi && is_vc1 == 0 ) // Raymond 2007/08/15
	free(mpi);
#endif
return ret;
}

#ifdef __ARC_MPLAYER__

// The decoder may have produced more than one frame when last called, in which case
// the remaining frames should be displayed first, before doing any more decoding
// (or at least be discarded in sync)
int decoded_frames_ready (void)
{
  //return the remaining number of frames to display
  if (mpvdec->frameInfo != NULL)
  {
    if (mpvdec->frameInfo->numFrames != 0)
      return (mpvdec->frameInfo->numFrames);
  }
  return (framesProduced);
}

void clear_frames_produced (void)
{
  if (mpvdec->frameInfo == NULL)
    framesProduced = 0;
}


#endif

