    #include <stdio.h>
    #include <errno.h>
    #include <stdlib.h>
    #include <assert.h>
    #include <time.h>
    #include <string.h>
    #include <fcntl.h>
    #include <sys/mman.h>
    
    
    #include "config.h"
    #include "mp_msg.h"
    #include "help_mp.h"
    
    #ifdef USE_LIBARCAVCODEC
    #include "bswap.h"
    
    #include "vd_internal.h"
    
    #include "arc_codecs.h"
    #include "../libarcavcodec/avcodec.h"
    #include "../libvo/arc_pgu.h"

    #include "libvo/video_out.h"	// Raymond - for HW scalar	
    
    //#define __ARC_NET__
    
    #ifdef __ARC_NET__
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #endif
    
    
    #define BUFSZ 1000000
    
    static vd_info_t info = {
    	"ARC subset of FFmpeg's libavcodec codec family",
    	"libarc",
    	"A'rpi",
    	"A'rpi, Michael, Alex",
    	"native codecs (http://ffmpeg.sf.net/)"
    };
    
    LIBVD_EXTERN(ffmpeg) // TODO: Should be libarc?
    
    extern arc_codec_functions mpg_functions;
    static arc_frame_data frameData;
    
    int avcodec_inited=0;
    
    typedef struct {
        AVCodecContext *avctx;
        AVFrame *pic;
        float last_aspect;
        int do_slices;
        int do_dr1;
        int vo_inited;
        int best_csp;
        int b_age;
        int ip_age[2];
        int qp_stat[32];
        double qp_sum;
        double inv_qp_sum;
        int ip_count;
        int b_count;
    } vd_ffmpeg_ctx;
    
    
    //#ifdef FF_POSTPROCESS
    //unsigned int lavc_pp=0;
    //#endif
    
    #include "m_option.h"
    
    #define NUM_BUFFERS	5 // XXX should NOT be here
    
    static unsigned long *virtualBufferAddresses;
    static unsigned long *physicalBufferAddresses;
    static int fdFrameDevice = -1;
    static int display_initialised = 0;
    static sh_video_t *local_sh = NULL;
    extern volatile void *memBase0;	//quick hack to get access to the virtual base address
    
    static int  InitOutput 		(void);
    static mp_image_t *CreateMpiImage (sh_video_t *);

// Raymond
static unsigned long *vDecBuffer;	// virtual
static unsigned long *pDecBuffer;	// physical
static int  InitDecBuffer(void);	

// Raymond 2007/04/24
unsigned int biCompression = 0;

    static int lavc_param_workaround_bugs= FF_BUG_AUTODETECT;
    static int lavc_param_error_resilience=2;
    static int lavc_param_error_concealment=3;
    static int lavc_param_gray=0;
    static int lavc_param_vstats=0;
    static int lavc_param_idct_algo=0;
    static int lavc_param_debug=0;
    static int lavc_param_vismv=0;
    static int lavc_param_skip_top=0;
    static int lavc_param_skip_bottom=0;
    static int lavc_param_fast=0;
    static int lavc_param_lowres=0;
    static char *lavc_param_lowres_str=NULL;
    
    static int sliceDataSize = 0;

    static int fdMem = -1;	// Raymond 2007/04/16
    unsigned long SDMBase = 0;
    
    #ifdef __ARC_NET__
    const char *socketAddress = "172.16.9.23";
    static int sd,cd,n,wsz,offset,abortNow = 0;
    static int socketEnabled = 0;
    static struct sockaddr_in sa;
    void transmit_arc (char *data,int sz);
    #endif
    
    
    static void SetSizes(int w,int h)
    {
      if (display_initialised == 0)
      {
        display_initialised = 1;
        fprintf(stderr,"SetSizes : Configuring display for %d x %d\n",w,h);
        local_sh->disp_h = h;
        local_sh->disp_w = w;
        mpcodecs_config_vo(local_sh,w,h,IMGFMT_YV12); //YUV420 format
        // Raymond 2008/04/11 - for MPEG-4 elementary stream
        if( use_DecBuffer == 1 )
	{
		if (!SetDecBufferResolution(w, h))
			return 0;
	}
        InitOutput();
	// Raymond 2008/04/11 - for MPEG-4 elementary stream	
	if( use_DecBuffer == 1 )	
		mpg_functions.set_buffer_addresses(SDMBase,vDecBuffer,pDecBuffer);
	else	
        mpg_functions.set_buffer_addresses(SDMBase,virtualBufferAddresses,physicalBufferAddresses);
      }
    }
    
    m_option_t lavc_decode_opts_conf[]={
    	{"bug", &lavc_param_workaround_bugs, CONF_TYPE_INT, CONF_RANGE, -1, 999999, NULL},
    	{"er", &lavc_param_error_resilience, CONF_TYPE_INT, CONF_RANGE, 0, 99, NULL},
    	{"gray", &lavc_param_gray, CONF_TYPE_FLAG, 0, 0, CODEC_FLAG_PART, NULL},
    	{"idct", &lavc_param_idct_algo, CONF_TYPE_INT, CONF_RANGE, 0, 99, NULL},
    	{"ec", &lavc_param_error_concealment, CONF_TYPE_INT, CONF_RANGE, 0, 99, NULL},
    	{"vstats", &lavc_param_vstats, CONF_TYPE_FLAG, 0, 0, 1, NULL},
    	{"debug", &lavc_param_debug, CONF_TYPE_INT, CONF_RANGE, 0, 9999999, NULL},
    	{"vismv", &lavc_param_vismv, CONF_TYPE_INT, CONF_RANGE, 0, 9999999, NULL},
    	{"st", &lavc_param_skip_top, CONF_TYPE_INT, CONF_RANGE, 0, 999, NULL},
    	{"sb", &lavc_param_skip_bottom, CONF_TYPE_INT, CONF_RANGE, 0, 999, NULL},
            {"fast", &lavc_param_fast, CONF_TYPE_FLAG, 0, 0, CODEC_FLAG2_FAST, NULL},
    	{"lowres", &lavc_param_lowres_str, CONF_TYPE_STRING, 0, 0, 0, NULL},
    	{NULL, NULL, 0, 0, 0, 0, NULL}
    };
    
    // to set/get/query special features/parameters
    static int control(sh_video_t *sh,int cmd,void* arg,...)
    {
      if (cmd == VDCTRL_RESYNC_STREAM)
        return (mpg_functions.control(cmd,arg));
      else
        return (CONTROL_UNKNOWN);
    }
    
    
    FILE *inputFile = NULL;
    
    // init driver
    static int init(sh_video_t *sh)
    {
//      int fdMem = -1;

	// Raymond - for HW scalar , 2006/10/26
	int dispw = sh->disp_w;
	int disph = sh->disp_h;
	
      local_sh = sh;
      
      mpg_functions.set_sizes = (fp_set_sizes)SetSizes;
      memset(&frameData,0,sizeof(arc_frame_data));
      
      //open a memory device to get access to the SDM memory
      fdMem = open("/dev/mem",O_RDWR|O_SYNC);
      if (fdMem < 0)
      {
          fprintf(stderr,"\nFailed to open /dev/mem");
          return (0);
      }
    
      SDMBase = (unsigned long)mmap((void *)0,32768,PROT_READ|PROT_WRITE,MAP_SHARED,fdMem,0xC0F00000);

	// Raymond 2004/04/16
	int ret = 0;
	if( HW_Scalar == 1 )
		ret = InitPhyMem(fdMem);
	
//      close(fdMem);
      if (SDMBase == (unsigned long)-1)
      {
        fprintf(stderr,"\nFailed on mmap [%s]",  strerror(errno));
        return (0);
      }

	// Raymond 2004/04/16
	if( HW_Scalar == 1 )
	{
		if (ret == 0)
      		{
       		fprintf(stderr,"\nFailed on InitPhyMem [%s]",  strerror(errno));
        		return (0);
      		}
	}	
	  
    #ifdef __ARC_NET__
      {
            int sz;
    	sd = socket(AF_INET,SOCK_STREAM,0);
    	if (sd == -1)
    	{
    	  fprintf(stderr,"\nUnable to create socket");
    	  return (2);
    	}
    	memset((void *)&sa,0,sizeof(struct sockaddr_in));
    	sa.sin_family = AF_INET;
    	sa.sin_addr.s_addr = inet_addr(socketAddress);
    	sa.sin_port = htons(1149);
    	int trycount = 0;
    	while (1)
    	{
    	  cd = connect(sd,(struct sockaddr *)&sa,sizeof(struct sockaddr));
    	  if (cd == -1)
    	  {
    	    fprintf(stderr,"\nUnable to connect!");
    	    sleep(2);
    	    trycount++;
    	    if (trycount == 5)
    	    {
    	      fprintf(stderr,"\nAborting!");
                  return (2);
    	    }
    	  }
    	  else
    	    break;
            }
            fprintf(stderr,"\nConnected to server\n");
      }
    #endif
      if ((sh->disp_w != 0) && (sh->disp_h != 0))
      {
#if 0 	//Barry 2007-04-27      
//#if 1      // Raymond 2006/10/02	
	      if (sh->disp_w & 0xf) 
			sh->disp_w = ((sh->disp_w + 0xf) & ~0xf);				  
		if (sh->disp_h & 0xf)
			sh->disp_h= ((sh->disp_h + 0xf) & ~0xf);
#endif

#if 0 	//Barry 2007-05-16
		// Raymond fix bug - 2006/11/16
		dispw = sh->disp_w;
		disph = sh->disp_h;
#endif

//        fprintf(stderr,"Configuring display for %d x %d\n",sh->disp_w,sh->disp_h);
        mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_YV12); //YUV420 format

	// Set decoder buffer resolution - Raymond 2006/12/27
	if( use_DecBuffer == 1 )
#if 0	//Barry 2007-05-28
		SetDecBufferResolution(dispw, disph);	
#else
	{
		if (!SetDecBufferResolution(dispw, disph))
			return 0;
	}
#endif
	
        InitOutput();
        display_initialised = 1;
      }
      int type = 0;
      if (!strcmp(sh->codec->dll,"mpegvideo"))
        type = ARC_MPEG1;
      if (!strcmp(sh->codec->dll,"mpeg2video"))
        type = ARC_MPEG2;
      if (!strcmp(sh->codec->dll,"mpeg4"))
        type = ARC_MPEG4;
      if (!strcmp(sh->codec->dll,"h263"))
        type = ARC_H263RAW;
      if (!strcmp(sh->codec->dll,"flv"))	//Barry 2007-10-31
        type = ARC_FLV;
      if (!strcmp(sh->codec->dll,"mjpeg"))
      {
        	type = ARC_MJPEG;

		// Raymond 2007/04/24
		biCompression = sh->bih->biCompression;
      	}

	// Raymond 2006/12/28
	if( use_DecBuffer == 1 )	
		mpg_functions.set_buffer_addresses(SDMBase,vDecBuffer,pDecBuffer);
	else	
		mpg_functions.set_buffer_addresses(SDMBase,virtualBufferAddresses,physicalBufferAddresses);

      //if (mpg_functions.init(type,NULL,0,0,sh->disp_w,sh->disp_h) == 0)
      if (mpg_functions.init(type,NULL,0,0,dispw,disph) == 0)// Raymond - for HW scalar , 2006/10/26	  	
      {
        fprintf(stderr,"\nFailed to initialise for codec %s format 0x%.8X",sh->codec->dll,sh->format);
        return (0);
      }
      return (1);
    }
    
    // uninit driver
    static void uninit(sh_video_t *sh)
    {
      mpg_functions.uninit();
      if (SDMBase != 0)
        munmap((void *)SDMBase,32768);

	// Raymond 2004/04/16
	if( HW_Scalar == 1 )
	  	ClosePhyMem();	

	if (fdMem >= 0)
		close(fdMem);

    #ifdef __ARC_NET__
        close(sd);
    #endif
    
    }
    
    // decode a frame
    static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags)
    {
      int numFrames = 0;
      int i;
      //fprintf(stderr,"\nDecode buffer address is 0x%.8X",data);
      if (len < 0)
      {
        //fprintf(stderr,"\nlen is %d",len);
        len = 0;
      }
      int ret = mpg_functions.decode(data,len,0,&numFrames,flags);
      if (numFrames == 0)
        return NULL;	// skipped image
    
      mp_image_t* mpi = CreateMpiImage (sh);
      if (!mpi->planes[0])
        return NULL;
    #ifdef __ARC_NET__
      if (mpi != NULL)
      {
        transmit_arc (mpi->planes[0],mpi->height*mpi->width);
        transmit_arc (mpi->planes[1],(mpi->height*mpi->width)/4);
        transmit_arc (mpi->planes[2],(mpi->height*mpi->width)/4);
      }
    #endif
      return (mpi);
    }
    
    static mp_image_t *CreateMpiImage (sh_video_t *sh)
    {
      mp_image_t * mpi = (mp_image_t *)malloc(sizeof(mp_image_t));
      if (mpi == NULL)
      {
        fprintf(stderr,"\nFailed in Malloc for mpi");
        return (NULL);
      }
      memset(mpi,0,sizeof(mp_image_t));
      mpi->flags = mpg_functions.get_frame_data(&frameData);
      mpi->flags |= MP_IMGFLAG_DIRECT | MP_IMGFLAG_PLANAR | MP_IMGFLAG_YUV;
      mpi->width = sh->disp_w;
      mpi->height = sh->disp_h;
#if 1	  
	//Barry 2007-04-27
       mpi->w = sh->disp_w;
       mpi->h = sh->disp_h;
#else
      mpi->w = frameData.width;
      mpi->h = frameData.height;
#endif	  
      mpi->x = frameData.xOffset;
      mpi->y = frameData.yOffset;
      mpi->stride[0] = frameData.stride[0];
      mpi->stride[1] = frameData.stride[1];
      mpi->stride[2] = frameData.stride[2];
      mpi->planes[0] = frameData.planes[0];
      mpi->planes[1] = frameData.planes[1];
      mpi->planes[2] = frameData.planes[2];
      mpi->num_planes = 3;
      mpi->priv = (void *)frameData.index;
      return (mpi);
    }
    
    static int InitOutput (void)
    {
      int i;
      short s;
      unsigned long physaddr;
      struct yuv_info yuv;
      //get the YUV buffer information
      fdFrameDevice = open("/dev/fb0",O_RDWR);
      if (fdFrameDevice == -1)
      {
        fprintf(stderr,"arcpgufb: can't open /dev/fb0: %s\n", strerror(errno));
        return -1;
      }
      if (ioctl(fdFrameDevice, ARCPGUFB_GETYUV, &yuv) == -1)
      {
        fprintf(stderr,"\nH264: ARCPGUFB_GETYUV ioctl failed due to %s", strerror(errno));
        return (0);
      }
      //and the related physical address information
      physaddr = yuv.phys_base;
      // fprintf(stderr,"\nPhysical base is 0x%X",physaddr);
      //check the number of buffers is correct
      if (yuv.num_buffers < NUM_BUFFERS)
      {
        fprintf(stderr,"\nNumber of driver buffers (%u) < %u",yuv.num_buffers, NUM_BUFFERS);
        return (0);
      }
      virtualBufferAddresses = malloc((yuv.num_buffers+1) * sizeof(void *));
      physicalBufferAddresses = malloc((yuv.num_buffers+1) * sizeof(long));
      //calculate the start addresses of all buffers
      for (i=0; i < yuv.num_buffers; i++)
      {
	int offset = ioctl(fdFrameDevice, ARCPGUFB_GET_YUV_BASE, i);
        virtualBufferAddresses[i] = (unsigned long)memBase0 + offset;
        physicalBufferAddresses[i] = physaddr + offset;
      }
      //set the tail data (used for range checking)
      virtualBufferAddresses[i] = virtualBufferAddresses[i-1] + yuv.yuv_size;
      physicalBufferAddresses[i] = physicalBufferAddresses[i-1] + yuv.yuv_size;

	// Raymond 2006/12/28
	if ( use_DecBuffer == 1 )
		InitDecBuffer ();
	  	
    }

static int InitDecBuffer (void)
{
	int i;
    	short s;
      	unsigned long physaddr;
      	struct yuv_info yuv;
      	//get the YUV buffer information
      	
#if 0
      	fdFrameDevice = open("/dev/fb0",O_RDWR);
      	if (fdFrameDevice == -1)
      	{
       	fprintf(stderr,"arcpgufb: can't open /dev/fb0: %s\n", strerror(errno));
        	return -1;
      	}
#endif		
      	if (ioctl(fdFrameDevice, ARCPGUFB_GET_DEC_YUV, &yuv) == -1)
      	{
       	fprintf(stderr,"\nARCPGUFB_GET_DEC_YUV ioctl failed due to %s", strerror(errno));
        	return (0);
      	}

//	printf("InitDecBuffer() : %d x %d\n", yuv.width, yuv.height);
	
      	//and the related physical address information
      	physaddr = yuv.phys_base;
      	// fprintf(stderr,"\nPhysical base is 0x%X",physaddr);
      	//check the number of buffers is correct
      	if (yuv.num_buffers < NUM_BUFFERS)
      	{
       	fprintf(stderr,"\nNumber of decoder buffers (%u) < %u",yuv.num_buffers, NUM_BUFFERS);
        	return (0);
      	}
      	vDecBuffer = malloc((yuv.num_buffers+1) * sizeof(void *));
      	pDecBuffer = malloc((yuv.num_buffers+1) * sizeof(long));
      	//calculate the start addresses of all buffers
      	for (i=0; i < yuv.num_buffers; i++)
      	{
		int offset = ioctl(fdFrameDevice, ARCPGUFB_GET_DEC_YUV_BASE, i);
        	vDecBuffer[i] = (unsigned long)memBase0 + offset;
        	pDecBuffer[i] = physaddr + offset;
      	}
      	//set the tail data (used for range checking)
      	vDecBuffer[i] = vDecBuffer[i-1] + yuv.yuv_size;
      	pDecBuffer[i] = pDecBuffer[i-1] + yuv.yuv_size;
}

    #ifdef __ARC_NET__
    void transmit_arc (char *data,int sz)
    {
        while (sz > 0)
        {
          int n = write(sd,data,wsz);
          if (n < 0)
          {
            fprintf(stderr,"\nerror on writing data [%s]",strerror(errno));
    	return;
          }
          sz -= n;
          data += n;
        }
    }
    #endif
    
    
    #endif
    
    
