
#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "mp_msg.h"
#include "stream.h"
#include "help_mp.h"
#include "../m_option.h"
#include "../m_struct.h"

#if 1	// Raymond 2007/07/02 - from Robert

#include <errno.h>

#define LOAD_ARGS_0()	\
	register int _r0  = (int) 0;
#define ASM_ARGS_0
#define LOAD_ARGS_1(r0)				\
  LOAD_ARGS_0 ()				\
	_r0  = (int) (r0);	
//#define ASM_ARGS_1	ASM_ARGS_0, "r" (_r0) 
#define ASM_ARGS_1	ASM_ARGS_0
#define LOAD_ARGS_2(r0, r1)			\
  register int _r1 asm ("r1") = (int) (r1);	\
  LOAD_ARGS_1 (r0)
#define ASM_ARGS_2	ASM_ARGS_1, "r" (_r1)

#define LOAD_ARGS_3(r0, r1, r2)			\
  register int _r2 asm ("r2") = (int) (r2);	\
  LOAD_ARGS_2 (r0, r1)
#define ASM_ARGS_3	ASM_ARGS_2, "r" (_r2)

#ifndef __set_errno
# define __set_errno(val) (*__errno_location ()) = (val)
#endif

#ifndef SYS_ify
//# define SYS_ify(syscall_name)	(__NR_##syscall_name)
#define SYS_ify(syscall_name)	225
#endif

#define INLINE_SYSCALL_NEW(name, nr, args...)				\
({ unsigned int _sys_result;						    \
    {								                 	\
	LOAD_ARGS_##nr (args)					        	\
									                    \
        asm volatile (							        \
		/* FIXME: do we need to save the sp here? */	\
		"mov r8, %1 \n\t"   /* save the syscall type in r8 */	\
		"mov r0, %2 \n \t" \
		"trap0 \n\t"  /* do the interrupt (syscall) */	        \
		"nop \n\t" /* two nops to flush pipeline (hw issue) */	\
		"nop \n\t"						                        \
        "mov %0, r0 \n\t"                                       \
	     : "=r" (_sys_result)						            \
	     : "i" (SYS_ify(name)), "r"(_r0) ASM_ARGS_##nr			        \
	     : "r0","r8");							                \
    }									                        \
    if(_sys_result >= (unsigned long) -125) {				    \
	        __set_errno(-_sys_result);			            	\
	        _sys_result = (unsigned long) -1;		        	\
    }								                        	\
	(int) _sys_result; })


#define _syscall_readphy(type,name,type1,arg1,type2,arg2,type3,arg3) \
type name(type1 arg1,type2 arg2,type3 arg3) \
{ \
return (type) (INLINE_SYSCALL_NEW(name, 3, arg1, arg2, arg3)); \
}

_syscall_readphy(ssize_t, readphy, int, fd, void *, buf, size_t, count);

int fill_buffer_phy(stream_t *s, char* buffer, int max_len){

  int r = readphy(s->fd,(void *)buffer,max_len);
  // Raymond 2007/04/13
  if( r < 0 )
  {  	
  	printf("fill_buffer_phy = -1\n");
	close(s->fd);	// Raymond 2007/04/20
	exit(-1);		// Raymond 2007/04/17
  }
//  printf("fill_buffer_phy = %d\n", r);
  return (r <= 0) ? -1 : r;
}

#endif

static struct stream_priv_s {
  char* filename;
} stream_priv_dflts = {
  NULL
};

#define ST_OFF(f) M_ST_OFF(struct stream_priv_s,f)
/// URL definition
static m_option_t stream_opts_fields[] = {
  {"filename", ST_OFF(filename), CONF_TYPE_STRING, 0, 0 ,0, NULL},
  { NULL, NULL, 0, 0, 0, 0,  NULL }
};
static struct m_struct_st stream_opts = {
  "file",
  sizeof(struct stream_priv_s),
  &stream_priv_dflts,
  stream_opts_fields
};  

// Raymond 2008/02/22
#include "uncached_buf.h"
extern int fill_size;	

// Raymond 2008/08/20
extern unsigned char * str_buf;	
#ifndef BUILD_MPLAYER_LITE	// Raymond 2008/08/26
extern int str_buf_size;
#endif
static int fill_buffer(stream_t *s, char* buffer, int max_len){

// Raymond 2008/02/22
if( fill_size > 0 )
{
#ifndef BUILD_MPLAYER_LITE	// Raymond 2008/08/26
	// Raymond 2008/08/20
	if( fill_size != str_buf_size )
	{
		sfree(str_buf);	
		str_buf = (unsigned short *)salloc(fill_size+512);
		str_buf_size = fill_size;
		buffer = s->buffer = str_buf + (512 -((int)(str_buf) & 511));
	}
#endif	
	unsigned int phy_buf = VIRTUAL_TO_UCM(buffer);	
	int flen =  fill_buffer_phy(s, (char *)phy_buf, max_len);
	uncached_flush_dcache( (void *) phy_buf, max_len);	// for SD
//	printf("fill_buffer_phy max_len = %d, fill_size = %d\n", max_len, fill_size);
	return flen;
}
else
{
  int r = read(s->fd,buffer,max_len);
  // Raymond 2007/04/13
  if( r < 0 )
  {  	
  	printf("fill_buffer = -1\n");
	close(s->fd);	// Raymond 2007/04/20
	exit(-1);		// Raymond 2007/04/17
  }
 
  return (r <= 0) ? -1 : r;
}  
}

static int write_buffer(stream_t *s, char* buffer, int len) {
  int r = write(s->fd,buffer,len);
  return (r <= 0) ? -1 : r;
}

static int seek(stream_t *s,off_t newpos) {
  s->pos = newpos;
  if(lseek(s->fd,s->pos,SEEK_SET)<0) {
    s->eof=1;
    return 0;
  }
  return 1;
}

static int seek_forward(stream_t *s,off_t newpos) {
  if(newpos<s->pos){
    mp_msg(MSGT_STREAM,MSGL_INFO,"Cannot seek backward in linear streams!\n");
    return 0;
  }
  while(s->pos<newpos){
    int len=s->fill_buffer(s,s->buffer,STREAM_BUFFER_SIZE);
    if(len<=0){ s->eof=1; s->buf_pos=s->buf_len=0; break; } // EOF
    s->buf_pos=0;
    s->buf_len=len;
    s->pos+=len;
  }
  return 1;
}

static int open_f(stream_t *stream,int mode, void* opts, int* file_format) {
  int f;
  mode_t m = 0;
  off_t len;
  struct stream_priv_s* p = (struct stream_priv_s*)opts;

  if(mode == STREAM_READ)
    m = O_RDONLY;
  else if(mode == STREAM_WRITE)
    m = O_WRONLY;
  else {
    mp_msg(MSGT_OPEN,MSGL_ERR, "[file] Unknown open mode %d\n",mode);
    m_struct_free(&stream_opts,opts);
    return STREAM_UNSUPORTED;
  }

  if(!p->filename) {
    mp_msg(MSGT_OPEN,MSGL_ERR, "[file] No filename\n");
    m_struct_free(&stream_opts,opts);
    return STREAM_ERROR;
  }

#if defined(__CYGWIN__)|| defined(__MINGW32__)
  m |= O_BINARY;
#endif    

  if(!strcmp(p->filename,"-")){
    if(mode == STREAM_READ) {
      // read from stdin
      mp_msg(MSGT_OPEN,MSGL_INFO,MSGTR_ReadSTDIN);
      f=0; // 0=stdin
#ifdef __MINGW32__
	  setmode(fileno(stdin),O_BINARY);
#endif
    } else {
      mp_msg(MSGT_OPEN,MSGL_INFO,"Writing to stdout\n");
      f=1;
#ifdef __MINGW32__
	  setmode(fileno(stdout),O_BINARY);
#endif
    }
  } else {
    f=open(p->filename,m);
    if(f<0) {
      mp_msg(MSGT_OPEN,MSGL_ERR,MSGTR_FileNotFound,p->filename);
      m_struct_free(&stream_opts,opts);
      return STREAM_ERROR;
    }
  }

  len=lseek(f,0,SEEK_END); lseek(f,0,SEEK_SET);
#ifdef __MINGW32__
  if(f==0 || len == -1) {
#else
  if(len == -1) {
#endif
    stream->seek = seek_forward;
    stream->type = STREAMTYPE_STREAM; // Must be move to STREAMTYPE_FILE
    stream->flags |= STREAM_SEEK_FW;
  } else if(len >= 0) {
    stream->seek = seek;
    stream->end_pos = len;
    stream->type = STREAMTYPE_FILE;
  }

#ifdef _LARGEFILE_SOURCE
  mp_msg(MSGT_OPEN,MSGL_V,"[file] File size is %lld bytes\n", (long long)len);
#else
  mp_msg(MSGT_OPEN,MSGL_V,"[file] File size is %u bytes\n", (unsigned int)len);
#endif 

  stream->fd = f;
  stream->fill_buffer = fill_buffer;
  stream->write_buffer = write_buffer;

  m_struct_free(&stream_opts,opts);
  return STREAM_OK;
}

stream_info_t stream_info_file = {
  "File",
  "file",
  "Albeu",
  "based on the code from ??? (probably Arpi)",
  open_f,
  { "file", "", NULL },
  &stream_opts,
  1 // Urls are an option string
};
