/*
 * Copyright (c) 2005 ARC International (UK) Ltd
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * May 2006
 */

#include "uncached_buf.h"
#include "arc_mem.h"
#include <stdio.h>
#include <stdlib.h>

void *baseVirtualAddress;
void *basePhysicalAddress = 0;
//int baseVirtualSize = 1024*1024;
int baseVirtualSize;	// Raymond 2008/01/09

int uncached_fd = 0;	// Raymond 2007/07/11

#if defined __ARC_LINUX__
#include <sys/mman.h>
#include <sys/errno.h>


char *uncached_bitstream_alloc(void)
{
    // Raymond 2007/07/11
    if(uncached_fd && basePhysicalAddress)
	return;
	
    const char *dev = "/dev/mem_uncached";
    void *vaddr;
    int fd = open(dev, 0666);

    if (basePhysicalAddress)
    {
	printf("uncached_bitstream_alloc called twice!\n");
	abort();
    }

    if (fd < 0)
	goto error;

    uncached_fd = fd;	// Raymond 2007/07/11

    if (ioctl(fd,  GET_LEN, &baseVirtualSize))
	goto error;
    
    if (ioctl(fd, GET_PHYS, &basePhysicalAddress))
	goto error;

    vaddr =  mmap(0, baseVirtualSize, PROT_READ|PROT_WRITE,  MAP_SHARED , fd, 0);
//    close(fd);	// Raymond 2007/07/11
    if ((int)vaddr == -1)
	goto error;

    baseVirtualAddress = vaddr;

    return vaddr;

error:
    perror(dev);
    fprintf(stderr, "Failed to map uncached input buffer\n");
    if (errno == ENOTTY)
	fprintf(stderr, "Please upgrade to a newer version of the linux Kernel\n");
    exit(1);
}

void uncached_bitstream_free(char *p)
{
    munmap(p, baseVirtualSize);
    baseVirtualAddress = NULL;
}

// Raymond 2007/07/11
void uncached_flush_dcache( void * start, int len)
{	
	invalidate_region_t ireg;	
	ireg.len = len;            	//physical data length	
	ireg.start = start;      //physical dram address
	//	printf("ireg: start = %X, len = %d\n", start, len);	
	if (ioctl(uncached_fd, FLUSH_DCACHE_REGION, &ireg))		
		printf("FLUSH_DCACHE_REGION error\n");
}

#else

char *uncached_bitstream_alloc(void)
{
    baseVirtualAddress = malloc(baseVirtualSize);
    if (!baseVirtualAddress)
    {
	perror("Failed to allocate bitstream input buffer.");
	exit(1);
    }
    return baseVirtualAddress;
}

void uncached_bitstream_free(char *p)
{
    free(p);
    baseVirtualAddress = NULL;
}

#endif
