#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <strings.h>
#include "wcfg_reader.h"

#define ERR_MSG printf
#define DBG_MSG printf


#ifndef true
#define true 1
#endif
#ifndef false
#define false 0
#endif

#ifdef WIN32
#define CONFIG_FILE_NAME "./wconf.sys"
#define WPA_CONF_ORG_FN	 "./wpa_supplicant.conf"
#define WPA_CONF_FN	     "./_wpa_supplicant.conf"
#define NET_CFG_FN       "./NetCfg"
#else
#define CONFIG_FILE_NAME "/mnt/sknand/wconf.sys"
#define WPA_CONF_ORG_FN	 "/etc/wpa_supplicant.conf"
#define WPA_CONF_FN	     "/tmp/conf/wpa_supplicant.conf"
#define NET_CFG_FN       "/tmp/conf/NetCfg"
#endif

NETNVRAM_t gNetCfg;

void UpdateNetCfg(char *tag, char *value)
{
    if(!strcmp(tag, "WIFI_MODE"))
    {
        gNetCfg.Mac_Config.Mode = atoi(value);
    }
    else if(!strcmp(tag, "WIFI_ENCR"))
    {
        gNetCfg.Mac_Config.Encr = atoi(value);        
    }
    else if(!strcmp(tag, "WIFI_AUTH"))
    {
        gNetCfg.Mac_Config.Auth = atoi(value);        
    }
    else if(!strcmp(tag, "WIFI_CHANNEL"))
    {
        gNetCfg.Mac_Config.Chan = atoi(value);        
    }
    else if(!strcmp(tag, "WIFI_SSID") && strlen(value))
    {
        strcpy(gNetCfg.Mac_Config.Ssid, value);
    }
    else if(!strcmp(tag, "WIFI_KEYINDEX"))
    {
        gNetCfg.Mac_Config.DefKey = atoi(value);        
    }
    else if(!strcmp(tag, "WIFI_KEY") && strlen(value))
    {
		int keyindex = gNetCfg.Mac_Config.DefKey;

        strcpy(gNetCfg.Mac_Config.Keys[keyindex], value);
    }
    else if(!strcmp(tag, "WIFI_PASSPH"))
    {
        strcpy(gNetCfg.Mac_Config.PassPh, value);
    }    
    else if(!strcmp(tag, "WIFI_IPSOURCE") && strlen(value))
    {
        gNetCfg.Tcp_Config.ip_source = atoi(value);	
    }
    else if(!strcmp(tag, "WIFI_IP"))
    {
        int k = 0;
        char * octetptr;
        octetptr = strtok(value,".");
        while(octetptr != NULL && k < 4)
		{
            gNetCfg.Tcp_Config.addr_ip[k] = atol(octetptr);
            k++;
            octetptr = strtok(NULL,".");
        }
    }
    else if(!strcmp(tag, "WIFI_SUBNET"))
    {
        int k = 0;
        char * octetptr;
        octetptr = strtok(value,".");
        while(octetptr != NULL && k < 4)
        {
            gNetCfg.Tcp_Config.addr_subnet[k] = atol(octetptr);
            k++;
            octetptr = strtok(NULL,".");
        }
    }
    else if(!strcmp(tag, "WIFI_GATEWAY"))
    {
        int k = 0;
        char * octetptr;
        octetptr = strtok(value,".");
        while(octetptr != NULL && k < 4)
        {
            gNetCfg.Tcp_Config.addr_gateway[k] = atol(octetptr);
            k++;
            octetptr = strtok(NULL,".");
        }
    }
    else if(!strcmp(tag, "WIFI_AUTOIPSTART"))
    {
        int k = 0;
        char * octetptr;
        octetptr = strtok(value,".");
        while(octetptr != NULL && k < 4)
        {
            gNetCfg.Tcp_Config.addr_start_autoip[k] = atol(octetptr);
            k++;
            octetptr = strtok(NULL,".");
	}
    }
    else if(!strcmp(tag, "WIFI_AUTOIPEND"))
    {
        int k = 0;
        char * octetptr;
        octetptr = strtok(value,".");
        while(octetptr != NULL && k < 4)
        {
            gNetCfg.Tcp_Config.addr_end_autoip[k] = atol(octetptr);
            k++;
            octetptr = strtok(NULL,".");
        }
    }
    else
    {
        DBG_MSG("Unknown WIFI Tag:%s With Value: %s \n", tag, value);
    }
}

#define RW_BUF_SIZE 200

unsigned int UpdataWpaCfg(void)
{
	char TempBuf[RW_BUF_SIZE];
	unsigned int Size;
	FILE* pWpaConfFileIn;
	FILE* pWpaConfFileOut;
	
	pWpaConfFileIn = fopen(WPA_CONF_ORG_FN,"rb");
	if(pWpaConfFileIn)
	{
		pWpaConfFileOut = fopen(WPA_CONF_FN,"wb");
		if(pWpaConfFileOut)
		{
			while((Size = fread(TempBuf,1,RW_BUF_SIZE,pWpaConfFileIn))> 0)
			{
				//DumpHex("wpa config file",TempBuf,Size);
				fwrite(TempBuf,1,Size,pWpaConfFileOut);
			}
			Size = 0;
			Size+= sprintf(&TempBuf[Size] ,"network={\n");
			Size+= sprintf(&TempBuf[Size] ,"ssid=\"%s\"\n",gNetCfg.Mac_Config.Ssid);

			
			switch(gNetCfg.Mac_Config.Encr)
			{
			case MAC_ENCR_NONE:
				Size+= sprintf(&TempBuf[Size] ,"key_mgmt=NONE\n");
				break;
			case MAC_ENCR_WEP64:
			case MAC_ENCR_WEP128:
				Size+= sprintf(&TempBuf[Size] ,"key_mgmt=NONE\n");
				Size+= sprintf(&TempBuf[Size] ,"wep_key%d=\"%s\"\n",gNetCfg.Mac_Config.DefKey,gNetCfg.Mac_Config.Keys[gNetCfg.Mac_Config.DefKey]);
				Size+= sprintf(&TempBuf[Size] ,"wep_tx_keyidx=%d\n",gNetCfg.Mac_Config.DefKey);
				break;
			case MAC_ENCR_WPAPSK_TKIP:
				Size+= sprintf(&TempBuf[Size] ,"key_mgmt=WPA-PSK\n");
				Size+= sprintf(&TempBuf[Size] ,"proto=WPA\n");
				Size+= sprintf(&TempBuf[Size] ,"pairwise=TKIP\n");
				Size+= sprintf(&TempBuf[Size] ,"psk=\"%s\"\n",gNetCfg.Mac_Config.PassPh);
				break;
			case MAC_ENCR_WPAPSK_AES:
				Size+= sprintf(&TempBuf[Size] ,"key_mgmt=WPA-PSK\n");
				Size+= sprintf(&TempBuf[Size] ,"proto=WPA\n");
				Size+= sprintf(&TempBuf[Size] ,"pairwise=CCMP\n");
				Size+= sprintf(&TempBuf[Size] ,"psk=\"%s\"\n",gNetCfg.Mac_Config.PassPh);
				break;
			case MAC_ENCR_WPA2PSK_TKIP:
				Size+= sprintf(&TempBuf[Size] ,"key_mgmt=WPA-PSK\n");
				Size+= sprintf(&TempBuf[Size] ,"proto=WPA2\n");
				Size+= sprintf(&TempBuf[Size] ,"pairwise=TKIP\n");
				Size+= sprintf(&TempBuf[Size] ,"psk=\"%s\"\n",gNetCfg.Mac_Config.PassPh);
				break;
			case MAC_ENCR_WPA2PSK_AES:
				Size+= sprintf(&TempBuf[Size] ,"key_mgmt=WPA-PSK\n");
				Size+= sprintf(&TempBuf[Size] ,"proto=WPA2\n");
				Size+= sprintf(&TempBuf[Size] ,"pairwise=CCMP\n");
				Size+= sprintf(&TempBuf[Size] ,"psk=\"%s\"\n",gNetCfg.Mac_Config.PassPh);
				break;
			}
			Size+= sprintf(&TempBuf[Size] ,"}\n\n");

			fwrite(TempBuf,1,Size,pWpaConfFileOut);

			fclose(pWpaConfFileOut);
		}
		else
			ERR_MSG("[ERR]JoinAP::fopen=%s",WPA_CONF_FN);

		fclose(pWpaConfFileIn);
	}
	else
		ERR_MSG("[ERR]JoinAP::fopen=%s",WPA_CONF_ORG_FN);
}


unsigned int GetWifiCfg(void)
{
    char* pWCfgBuf;
    char linebuf[128];
    char tag[128];
    char value[128];
    int linelen,taglen,WCfgFileSize;
    FILE* pWCfgFile;
    char *ptr;
    char *ptr1;
    char *ptr2;
	unsigned int lRet = false;

    if((pWCfgFile = fopen(CONFIG_FILE_NAME, "rb")))
    {
		fseek(pWCfgFile,0,SEEK_END);
		WCfgFileSize = ftell(pWCfgFile);
		fseek(pWCfgFile,0,SEEK_SET);

		pWCfgBuf = malloc(WCfgFileSize+1);

		if(pWCfgBuf)
		{
			if(fread(pWCfgBuf, sizeof(char), WCfgFileSize,pWCfgFile)==WCfgFileSize)
			{
				memset(&gNetCfg, 0 ,sizeof(NETNVRAM_t));
				
				pWCfgBuf[WCfgFileSize] = '\0';

				ptr = pWCfgBuf;

				while(ptr1 = strchr(ptr, ';'))
				{
					memset(tag, 0, 128);
					memset(value, 0, 128);
					linelen = ptr1 - ptr;
					strncpy(linebuf, ptr, linelen);
					linebuf[linelen] = '\0';

					ptr = ptr1+ 3;//strlen(";\r\n");
					ptr2 = strchr(linebuf, '=');
					taglen = ptr2 - linebuf;
					strncpy(tag, linebuf, taglen);
					strcpy(value, ptr2+1);
					UpdateNetCfg(tag, value);
				}
				lRet = true;
			}
			else
				ERR_MSG("[ERR]GetWifiCfg::fread::EOF\n");

			free(pWCfgBuf);
		}
		else
			ERR_MSG("[ERR]GetWifiCfg::malloc\n");

		fclose(pWCfgFile);
    }
    else
		ERR_MSG("[ERR]GetWifiCfg::fopen::%s\n",CONFIG_FILE_NAME);
    	
    return lRet;
}

unsigned int FileExist(char* fn)
{
	unsigned int lRet = false;
	FILE* pFile;
	
	if(fn)
	{
		pFile = fopen(fn,"r");		
		if(pFile)
		{
			lRet = true;
			fclose(pFile);
		}
	}
	return lRet;
}

#define IS_INTERFACE_UP() FileExist("/var/run/udhcpc.rausb0.pid")

unsigned int SetIP(char* pIP,char* pSubNet ,char* pGW)
{
	pid_t pid;
	int status;


	if(IS_INTERFACE_UP())
	{
		pid = fork();
    	
		switch(pid)
		{
			case -1:
				perror("Fork failed!");
				break;
			case 0:
				//child process
				DBG_MSG("[RTMSG]SetIP::ifdown\n");
				execl("/sbin/ifdown", "ifdown","rausb0", NULL);
				_exit;
			default:
				//parent process
				waitpid(pid, &status, 0);
				
				DBG_MSG("[RTMSG]SetIP::ifdown::done\n");
		}
	}
	
	pid = fork();

	switch(pid)
	{
		case -1:
			perror("Fork failed!");
			break;
		case 0:
			DBG_MSG("[RTMSG]SetIP::ifconfig\n");
			execl("/sbin/ifconfig", "ifconfig","rausb0",pIP,"netmask",pSubNet,NULL);					
			_exit;
		default:
			//parent process
			waitpid(pid, &status, 0);
			DBG_MSG("[RTMSG]SetIP::ifconfig::done\n");

			pid = fork();

			switch(pid)
			{
			case -1:
				perror("Fork failed!");
				break;
			case 0:
				//child process
				DBG_MSG("[RTMSG]SetIP::route\n");
				execl("/sbin/route","route","add","default","gw",pGW, NULL);
				_exit;
			default:
				//parent process
				waitpid(pid, &status, 0);										
				DBG_MSG("[RTMSG]SetIP::route::done\n");
				return true;
			}			
	}
				
	return false;
}

int main(void)
{
	FILE* NetCfgFile;
	char strIP[20],strGW[20],strSubNet[20];

	//
	// Load NetCfg from file
	// 
	if((NetCfgFile = fopen(NET_CFG_FN, "rb")))
	{
		if(fread(&gNetCfg,sizeof(unsigned char),sizeof(NETNVRAM_t),NetCfgFile)!=sizeof(NETNVRAM_t))
		{
			memset(&gNetCfg, 0 ,sizeof(NETNVRAM_t));
		}
		fclose(NetCfgFile);
	}
	else
	{
		DBG_MSG("[DBG]Wcfg_reader::fopen::%s\n",NET_CFG_FN);
		memset(&gNetCfg, 0 ,sizeof(NETNVRAM_t));
	}
	
	if(GetWifiCfg())
	{
		UpdataWpaCfg();

		//
		// Save NetCfg to file
		//

		if((NetCfgFile = fopen(NET_CFG_FN, "wb")))
		{
			fwrite(&gNetCfg,sizeof(unsigned char),sizeof(NETNVRAM_t),NetCfgFile);
			fclose(NetCfgFile);
		}
	}
	
	if(gNetCfg.Tcp_Config.addr_ip[0] != 0 &&
		gNetCfg.Tcp_Config.addr_ip[1] != 0 &&
		gNetCfg.Tcp_Config.addr_ip[2] != 0 &&
		gNetCfg.Tcp_Config.addr_ip[3] != 0)
	{
#if 0		
		_itoa(gNetCfg.Tcp_Config.addr_ip    		,strIP		,20);
		_itoa(gNetCfg.Tcp_Config.addr_subnet		,strSubNet	,20);
		_itoa(gNetCfg.Tcp_Config.addr_gateway    	,strIP		,20);
#else
		sscanf(strIP	,"%d.%d.%d.%d",
			gNetCfg.Tcp_Config.addr_ip[0],
			gNetCfg.Tcp_Config.addr_ip[1],
			gNetCfg.Tcp_Config.addr_ip[2],
			gNetCfg.Tcp_Config.addr_ip[3]);
		DBG_MSG("[DBG]Wcfg_reader::IP=%s\n",strIP);	
		
		sscanf(strSubNet,"%d.%d.%d.%d",
			gNetCfg.Tcp_Config.addr_subnet[0],
			gNetCfg.Tcp_Config.addr_subnet[1],
			gNetCfg.Tcp_Config.addr_subnet[2],
			gNetCfg.Tcp_Config.addr_subnet[3]);		
		DBG_MSG("[DBG]Wcfg_reader::NetMask=%s\n",strSubNet);	
		
		sscanf(strGW	,"%d.%d.%d.%d",
			gNetCfg.Tcp_Config.addr_gateway[0],
			gNetCfg.Tcp_Config.addr_gateway[1],
			gNetCfg.Tcp_Config.addr_gateway[2],
			gNetCfg.Tcp_Config.addr_gateway[3]);
		DBG_MSG("[DBG]Wcfg_reader::Gateway=%s\n",strGW);	
		
#endif		

		SetIP(strIP,strSubNet,strIP);

	}
	
	exit(1);
}
